function crossValidation( V,V_rev,labels,folds, name )
% Evaluates the WTF detector using 5-fold cross-validation.

tmp_model = 'tmp.yml';	% Path to save the model temporarily

labels_cell = cell(5,1);
scores_cell = cell(5,1);

for f = unique(folds)'
    
    train_id = folds~=f;
    evl_id = folds==f;
    
    % libWTF (C++)
    train(...
        [V(train_id,:);V_rev(train_id,:)],...
        [labels(train_id);labels(train_id)],...
        tmp_model, 25);
    scores = classifyMultiple(V(evl_id,:),V_rev(evl_id,:),tmp_model);
    
    % Uncomment to use Matlab's AdaBoost
    %model = trainMatlab([V(train_id,:);V_rev(train_id,:)],...
    %    [labels(train_id);labels(train_id)], 100);
    %scores = classifyMultipleMatlab(V(evl_id,:),V_rev(evl_id,:),model);

    labels_cell{f} = labels(evl_id);
    scores_cell{f} = scores;
end
delete(tmp_model)

[mfpr, mtpr, mauc, mfpr99] = mean_roc_folds(scores_cell, labels_cell, 0.99);

figure('Color', 'white');
plot(mfpr, mtpr, 'LineWidth', 2, 'MarkerSize', 8);
hold on;
plot([0 1], [0.99 0.99], '--k', 'LineWidth', 2);
 
xlabel('False Positive Rate', 'FontSize', 20, 'FontName', 'Times');
ylabel('True Positive Rate', 'FontSize', 20, 'FontName', 'Times');

set(gca, 'YTick', [0.91:0.02:0.97 0.99 1]);

set(gca, 'FontSize', 20);
set(gca, 'FontName', 'Times');
set(gca, 'LineWidth', 2);

ylim([0.9 1]);
xlim([0 1]);

grid;
hl = legend(sprintf('%s (%.03f f99, %.03f AUC)', name, mfpr99, mauc), ...
    'Location', 'SouthEast');
set(hl, 'FontSize', 16);
end

function model = trainMatlab(V, labels, iters)
    model = fitensemble(V, double(labels), 'AdaBoostM1', iters, 'Tree');
end

function scores = classifyMultipleMatlab(V, V_rev, model)
    [~, dv] = predict(model, V);
    dv = dv(:,2);
    
    [~, dv_rev] = predict(model, V_rev);
    dv_rev = dv_rev(:,2);
    scores = min(dv, dv_rev);
end

function [fpr, tpr, auc, fpr_at_r] = mean_roc_folds(scores, labels, r)
    % Computes a receiver operating characteristic (ROC) curve and the area under the curve (AUC) based on
    % cross-validation results.
    %
    % Inputs:
    %   scores: (F x 1) cell array of the classifier scores of the feature vectors in each fold. (F is the number of
    %           folds.) (This can be e.g. the 'decision_values' output argument of wtf_adaboost.)
    %   labels: (F x 1) cell array of binary vectors of the ground truth labels of each feature vector. (N is the total number of feature vectors.)
    %   folds:  (N x 1) vector of the fold ids of each feature vector (1-based).
    %   r:      If specified, will output the mean false positive rate at true positive rate r.
    %
    
    % Compute the ROC curve using "vertical averaging", i.e. averaging tpr values for each fpr.
    %[fpr, tpr, ~, auc] = perfcurve(labels_cell, scores, 1, 'XVals', 'all', 'Xcrit', 'FPR', 'Ycrit', 'TPR');
    
    % With subsampling and interpolation for aesthetic marker distances in plots.
    [fpr, tpr, ~, ~] = perfcurve(labels, scores, 1, 'XVals', 0:1/25:1, 'Xcrit', 'FPR', 'Ycrit', 'TPR');
    
    % Without subsampling for accurate values.
    [fpr_full, tpr_full, ~, auc] = perfcurve(labels, scores, 1, 'XVals', 'all', 'Xcrit', 'FPR', 'Ycrit', 'TPR');
    
    tpr = tpr(:,1); % Remove confidence bounds
    tpr_full = tpr_full(:,1);
    
    auc = auc(1);
    fpr_at_r = fpr_full(find(tpr_full <= r, 1, 'last'));
end
