#include "extract_feature_vector.h"
#include "util.h"


#include <string>
#include <opencv2/opencv.hpp>
#include <cmath>

using namespace std;
using namespace cv;

namespace wtf {

/*******************************
 * Public functions in libwtf.h
 *******************************/
bool extract_feature_vector(
    const vector<double> &homography,
    const double width1,
    const double height1,
    const double width2,
    const double height2,
    const HistogramShape histogram_shape,
    const int n_histogram_bins,
    const vector<double> &x1s,
    const vector<double> &y1s,
    const vector<double> &s1s,
    const vector<double> &x2s,
    const vector<double> &y2s,
    const vector<double> &s2s,
    pair<vector<float>,vector<float> > &feature_vector ) 
{
    vector<float> fv1, fv2;

    Mat H, H_inv;
    try {
        H = vec2homography(homography);
    } catch (string e) {
        cout << e << endl;
        return false;
    }

    // Check singularity of H 
    if (determinant(H)==0) {
        return false;
    }    

    H_inv = H.inv();
        
    if ( !extract_feature_vector(
        H,width1,height1,width2,height2,
        histogram_shape,n_histogram_bins,
        x1s, y1s, s1s, fv1) |
        !extract_feature_vector(
        H_inv,width2,height2,width1,height1,
        histogram_shape,n_histogram_bins,
        x2s, y2s, s2s, fv2) ) { return false; }
        
    feature_vector = pair<vector<float>,vector<float> >(fv1,fv2);
    
    return true;
}

bool extract_feature_vector(
    const vector<double> &homography,
    const double width1,
    const double height1,
    const double width2,
    const double height2,
    const HistogramShape histogram_shape,
    const int n_histogram_bins,
    const vector<double> &x1s,
    const vector<double> &y1s,
    const vector<double> &a1s,
    const vector<double> &b1s,
    const vector<double> &c1s,
    const vector<double> &x2s,
    const vector<double> &y2s,
    const vector<double> &a2s,
    const vector<double> &b2s,
    const vector<double> &c2s,
    pair<vector<float>,vector<float> > &feature_vector ) 
{
    // Check length of as, bs, cs 
    int n_a1 = a1s.size();
    if (n_a1!=b1s.size()|n_a1!=c1s.size()) { return false; }
    int n_a2 = a2s.size();
    if (n_a2!=b2s.size()|n_a2!=c2s.size()) { return false; }
    
    // Transfer a,b,c into s
    vector<double> s1s(n_a1);
    for (int i=0; i<n_a1; i++) {
        s1s[i] = eclipse_scale(a1s[i],b1s[i],c1s[i]);
    }
    vector<double> s2s(n_a2);
    for (int i=0; i<n_a2; i++) {
        s2s[i] = eclipse_scale(a2s[i],b2s[i],c2s[i]);
    }

    return extract_feature_vector (
        homography, width1, height1, width2, height2,
        histogram_shape, n_histogram_bins,
        x1s, y1s, s1s, x2s, y2s, s2s,
        feature_vector );
}
/* END of public functions in libwtf.h 
 **************************************/


/****************************************
 * Functions in extract_feature_vector.h 
 ****************************************/

bool extract_feature_vector(
    const cv::Mat &homography,
    const double width1,
    const double height1,
    const double width2,
    const double height2,
    const HistogramShape histogram_shape,
    const int n_histogram_bins,
    const vector<double> &xs,
    const vector<double> &ys,
    const vector<double> &ss,
    vector<float> &features ) 
{
//    cout << homography <<endl;

    Mat overlap, inliers;
    
    // Compute the overlapping region 
    if ( !overlap_map(homography,width1,height1,width2,height2,overlap) ) {
        return false;
    }
    // Compute the inliers map
    if ( !inlier_map(width1,height1,xs,ys,ss,inliers) ) {
        return false;
    }

    if (histogram_shape==DIST2BORDER) {
        features = histogram_dist2border(overlap,inliers,n_histogram_bins);
    } else if (histogram_shape==DIST2CENTER) {
        features = histogram_dist2center(overlap,inliers,n_histogram_bins);
    } else {
        return false;
    }
    
    inliers &= overlap;
    features.push_back((float)countNonZero(inliers)/(float)inliers.size().area());
    
    return true;
}

vector<float> histogram_dist2center(
    const Mat &overlap,
    const Mat &inliers,
    const int n_histogram_bins )
{
    vector<float> h(n_histogram_bins);

    const Size size = overlap.size();
    const double diag = sqrt(
        size.width*size.width+size.height*size.height);
    const double dr = diag/n_histogram_bins/2;  // delta radius
    const Point2d center((double)size.width/2,(double)size.height/2);
    Mat bin_region;

    for (int i=0; i<n_histogram_bins; i++) {
        bin_region = Mat(size,CV_8U,Scalar(0));
        
        // Draw a ring
        circle(bin_region,center,dr*(i+1),1,-1);
        
        if (i!=0) {
            circle(bin_region,center,dr*(i),0,-1);
        }
        
        bin_region &= overlap;
        int normalize_factor = countNonZero(bin_region);
        
        if (normalize_factor==0) {
            h[i] = 0;
            continue;
        }
        
        bin_region &= inliers;
        int num_pixels = countNonZero(bin_region);
        
        h[i] = (float)num_pixels/(float)normalize_factor;
        
//        imshow("bin_region "+std::to_string(i),bin_region*255);
    }
    return h;
}

vector<float> histogram_dist2border(
    const Mat &overlap,
    const Mat &inliers,
    const int n_histogram_bins )
{
    vector<float> h(n_histogram_bins);
    
    const Size size = overlap.size();
    const int width = size.width;
    const int height = size.height;
    const int dx = width/2/n_histogram_bins;
    const int dy = height/2/n_histogram_bins;
    const int d = (dx<dy)? dx : dy;
    Rect rect(0,0,0,0);
    Mat bin_region;
    
//    cout << "dx, dy, d = " << dx <<" "<<dy<<" "<<d <<endl;
    
    for (int i=0; i<n_histogram_bins; i++) {
    
        if (i==0) {
            bin_region = Mat(size,CV_8U,Scalar(1));
        } else {
            bin_region = Mat(size,CV_8U,Scalar(0));
        }
        
        // Reuse the bigger rectangle from previous iteration.
        rectangle(bin_region,rect,Scalar(1),-1);
//        rect = Rect(dx*(i+1),dy*(i+1),dx*(n_histogram_bins-i-1)*2,dy*(n_histogram_bins-i-1)*2);
        int d_2border = d*(i+1);
        rect = Rect(
            d_2border,d_2border,
            width-d_2border*2,height-d_2border*2);
        rectangle(bin_region,rect,Scalar(0),-1);

//        imshow("bin_region "+std::to_string(i),bin_region*255);

        bin_region &= overlap;
        int normalize_factor = countNonZero(bin_region);
        
        if (normalize_factor==0) {
            h[i] = 0;
            continue;
        }
        
        bin_region &= inliers;
        int num_pixels = countNonZero(bin_region);
        
        h[i] = (float)num_pixels/(float)normalize_factor;
    }
    
    return h;
}

bool inlier_map(
    const double width,
    const double height,
    const vector<double> &xs,
    const vector<double> &ys,
    const vector<double> &ss,
    Mat &output )
{
    int n_inliers = xs.size();
    if (ys.size()!=n_inliers||ss.size()!=n_inliers) {
        return false;
    }

    // Create an empty image
    output = Mat(height,width,CV_8U,Scalar(0));
    
    for (int i=0; i<n_inliers; i++) {
        // Draw a circle 
        circle(output,Point2d(xs[i],ys[i]),ss[i],1,-1);
    }
//    imshow("Inliers Image", output*255);
    return true;
}

bool overlap_map(
    const Mat &homography,
    const int width1, 
    const int height1,
    const int width2, 
    const int height2,
    Mat &output )
{
//     cout << homography <<endl;

    output = Mat(height1,width1,CV_8U);

    Mat input = Mat::ones(Size(width2,height2),CV_8U);

    warpPerspective(input,output,homography,output.size(),INTER_NEAREST);

//    imshow("Overlap Image", output*255);
    return true;
}

}  // namespace wtf

